from __future__ import annotations

import ckan.plugins.toolkit as tk
from ckan.logic import NotFound

import ckanext.scheming.helpers as sch


def get_relations_info(pkg_type: str) -> list[tuple[str, str, str]]:
    """Return information about relation (related_entity, related_entity_type,
    relation_type) of specified package type (pkg_type) from schema.

    Returns:
        List of tuples of related entities: entity, entity_type, relation_type.
    """
    schema = sch.scheming_get_schema("dataset", pkg_type)
    if not schema:
        return []
    return [
        (
            field["related_entity"],
            field["related_entity_type"],
            field["relation_type"],
        )
        for field in schema["dataset_fields"]
        if "relationship_related_entity" in field.get("validators", "")
    ]


def get_relation_field(
    pkg_type: str,
    object_entity: str,
    object_entity_type: str,
    relation_type: str,
) -> dict[str, str]:
    """Return field dict for specified package type (pkg_type) describes relation
    with specified entity (object_entity, object_entity_type) and type of relation
    (relation_type).
    """
    schema = sch.scheming_get_schema("dataset", pkg_type)
    if not schema:
        return []
    for field in schema["dataset_fields"]:
        if (
            field.get("related_entity") == object_entity
            and field.get("related_entity_type") == object_entity_type
            and field.get("relation_type") == relation_type
        ):
            return field
    return {}


def entity_name_by_id(entity_id):
    """
    Returns pkg name by its id
    """

    try:
        pkg = tk.get_action("package_show")({"ignore_auth": True}, {"id": entity_id})
        if pkg:
            return pkg.get("name")
    except NotFound:
        pass

    try:
        org = tk.get_action("organization_show")(
            {"ignore_auth": True}, {"id": entity_id}
        )
        if org:
            return org.get("name")
    except NotFound:
        pass

    try:
        group = tk.get_action("group_show")({"ignore_auth": True}, {"id": entity_id})
        if group:
            return group.get("name")
    except NotFound:
        pass



from __future__ import annotations

from sqlalchemy import or_

import ckan.logic as logic
import ckan.plugins.toolkit as tk
from ckan.logic import validate

import ckanext.relationship.logic.schema as schema
from ckanext.relationship.model.relationship import Relationship
from ckanext.relationship.utils import entity_name_by_id

NotFound = logic.NotFound


@validate(schema.relation_create)
def relationship_relation_create(context, data_dict) -> list[dict[str, str]]:
    """Create relation with specified type (relation_type) between two entities
    specified by ids (subject_id, object_id). Also create reverse relation."""
    tk.check_access("relationship_relation_create", context, data_dict)

    subject_id = data_dict["subject_id"]
    object_id = data_dict["object_id"]
    relation_type = data_dict.get("relation_type")

    if Relationship.by_object_id(subject_id, object_id, relation_type):
        return

    relation = Relationship(
        subject_id=subject_id, object_id=object_id, relation_type=relation_type
    )

    reverse_relation = Relationship(
        subject_id=object_id,
        object_id=subject_id,
        relation_type=Relationship.reverse_relation_type[relation_type],
    )

    context["session"].add(relation)
    context["session"].add(reverse_relation)
    context["session"].commit()

    return [rel.as_dict() for rel in (relation, reverse_relation)]


@validate(schema.relation_delete)
def relationship_relation_delete(context, data_dict) -> list[dict[str, str]]:
    """Delete relation with specified type (relation_type) between two entities
    specified by ids (subject_id, object_id). Also delete reverse relation."""
    tk.check_access("relationship_relation_delete", context, data_dict)

    subject_id = data_dict["subject_id"]
    subject_name = entity_name_by_id(data_dict["subject_id"])
    object_id = data_dict["object_id"]
    object_name = entity_name_by_id(data_dict["object_id"])
    relation_type = data_dict.get("relation_type")

    relation = (
        context["session"]
        .query(Relationship)
        .filter(
            or_(
                Relationship.subject_id == subject_id,
                Relationship.subject_id == subject_name,
            ),
            or_(
                Relationship.object_id == object_id,
                Relationship.object_id == object_name,
            ),
        )
    )

    if relation_type:
        relation = relation.filter(Relationship.relation_type == relation_type)

    relation = relation.all()

    reverse_relation = (
        context["session"]
        .query(Relationship)
        .filter(
            or_(
                Relationship.subject_id == object_id,
                Relationship.subject_id == object_name,
            ),
            or_(
                Relationship.object_id == subject_id,
                Relationship.object_id == subject_name,
            ),
        )
    )

    if relation_type:
        reverse_relation = reverse_relation.filter(
            Relationship.relation_type
            == Relationship.reverse_relation_type[relation_type]
        )

    reverse_relation = reverse_relation.all()

    [context["session"].delete(rel) for rel in relation]
    [context["session"].delete(rel) for rel in reverse_relation]
    context["session"].commit()
    return [rel[0].as_dict() for rel in (relation, reverse_relation) if len(rel) > 0]


@validate(schema.relations_list)
def relationship_relations_list(context, data_dict) -> list[dict[str, str]]:
    """Return dicts list of relation of specified entity (object_entity, object_type)
    related with specified type of relation (relation_type) with entity specified
    by id (subject_id).
    """
    tk.check_access("relationship_relations_list", context, data_dict)

    subject_id = data_dict["subject_id"]
    object_entity = data_dict.get("object_entity")
    object_entity = (
        "group" if object_entity and object_entity == "organization" else object_entity
    )
    object_type = data_dict.get("object_type")
    relation_type = data_dict.get("relation_type")

    relations = Relationship.by_subject_id(
        subject_id, object_entity, object_type, relation_type
    )
    if not relations:
        return []
    return [rel.as_dict() for rel in relations]


@validate(schema.relations_ids_list)
def relationship_relations_ids_list(context, data_dict) -> list[str]:
    """Return ids list of specified entity (object_entity, object_type) related
    with specified type of relation (relation_type) with entity specified
    by id (subject_id).
    """
    tk.check_access("relationship_relations_ids_list", context, data_dict)

    rel_list = relationship_relations_list(context, data_dict)

    return list(set([rel["object_id"] for rel in rel_list]))


@validate(schema.get_entity_list)
def relationship_get_entity_list(context, data_dict) -> list[str]:
    """Return ids list of specified entity (entity, entity_type)"""
    tk.check_access("relationship_get_entity_list", context, data_dict)

    model = context["model"]

    entity = data_dict["entity"]
    entity = entity if entity != "organization" else "group"

    entity_type = data_dict["entity_type"]

    entity_class = logic.model_name_to_class(model, entity)

    entity_list = (
        context["session"]
        .query(entity_class.id, entity_class.name, entity_class.title)
        .filter(entity_class.state != "deleted")
        .filter(entity_class.type == entity_type)
        .all()
    )

    return entity_list



from __future__ import annotations

import json

import ckan.plugins.toolkit as tk


def relationship_get_entity_list(entity, entity_type, include_private=True):
    """Return ids list of specified entity (entity, entity_type)"""
    context = {}
    if entity == "package":
        entity_list = tk.get_action("package_search")(
            context,
            {
                "fq": f"type:{entity_type}",
                "fl": "id, name, title",
                "rows": 1000,
                "include_private": include_private,
            },
        )
        entity_list = entity_list["results"]
    else:
        entity_list = tk.get_action("relationship_get_entity_list")(
            context, {"entity": entity, "entity_type": entity_type}
        )
        entity_list = [
            {"id": id, "name": name, "title": title} for id, name, title in entity_list
        ]
    return entity_list


def relationship_get_current_relations_list(data, field) -> list[str]:
    """Pull existing relations for form_snippet and display_snippet."""
    subject_id = field.get("id")
    subject_name = field.get("name")
    if not subject_id and not subject_name:
        return []
    related_entity = data["related_entity"]
    related_entity_type = data["related_entity_type"]
    relation_type = data["relation_type"]

    current_relation_by_id = []
    current_relation_by_name = []

    if subject_id:
        current_relation_by_id = tk.get_action("relationship_relations_ids_list")(
            {},
            {
                "subject_id": subject_id,
                "object_entity": related_entity,
                "object_type": related_entity_type,
                "relation_type": relation_type,
            },
        )
    if subject_name:
        current_relation_by_name = tk.get_action("relationship_relations_ids_list")(
            {},
            {
                "subject_id": subject_name,
                "object_entity": related_entity,
                "object_type": related_entity_type,
                "relation_type": relation_type,
            },
        )
    return current_relation_by_id + current_relation_by_name


def relationship_get_selected_json(selected_ids: list = []) -> str:
    selected_pkgs = []
    for pkg_id in selected_ids:
        try:
            pkg_dict = tk.get_action("package_show")({}, {"id": pkg_id})
            selected_pkgs.append({"name": pkg_dict["id"], "title": pkg_dict["title"]})
        except Exception:
            continue
    return json.dumps(selected_pkgs)



from sqlalchemy import Column, Text, or_

import ckan.logic as logic
import ckan.model as model
from ckan.model.types import make_uuid

from .base import Base


class Relationship(Base):
    __tablename__ = "relationship_relationship"
    id: str = Column(Text, primary_key=True, default=make_uuid)
    subject_id: str = Column(Text, nullable=False)
    object_id: str = Column(Text, nullable=False)
    relation_type: str = Column(Text, nullable=False)

    reverse_relation_type = {
        "related_to": "related_to",
        "child_of": "parent_of",
        "parent_of": "child_of",
    }

    def __repr__(self):
        return (
            "Relationship("
            f"id={self.id!r},"
            f"subject_id={self.subject_id!r},"
            f"object_id={self.object_id!r},"
            f"relation_type={self.relation_type!r},"
            ")"
        )

    def as_dict(self):
        id = self.id
        subject_id = self.subject_id
        object_id = self.object_id
        relation_type = self.relation_type
        return {
            "id": id,
            "subject_id": subject_id,
            "object_id": object_id,
            "relation_type": relation_type,
        }

    @classmethod
    def by_object_id(cls, subject_id, object_id, relation_type):
        subject_name = _entity_name_by_id(subject_id)
        object_name = _entity_name_by_id(object_id)

        return (
            model.Session.query(cls)
            .filter(
                or_(
                    cls.subject_id == subject_id,
                    cls.subject_id == subject_name,
                )
            )
            .filter(or_(cls.object_id == object_id, cls.object_id == object_name))
            .filter(cls.relation_type == relation_type)
            .one_or_none()
        )

    @classmethod
    def by_subject_id(
        cls, subject_id, object_entity, object_type=None, relation_type=None
    ):
        subject_name = _entity_name_by_id(subject_id)

        q = model.Session.query(cls).filter(
            or_(cls.subject_id == subject_id, cls.subject_id == subject_name)
        )

        if object_entity:
            object_class = logic.model_name_to_class(model, object_entity)
            q = q.filter(
                or_(
                    object_class.id == cls.object_id,
                    object_class.name == cls.object_id,
                )
            )

            if object_type:
                q = q.filter(object_class.type == object_type)

        if relation_type:
            q = q.filter(cls.relation_type == relation_type)

        return q.distinct().all()


def _entity_name_by_id(entity_id):
    """
    Returns entity (package or organization or group) name by its id
    """

    pkg = (
        model.Session.query(model.Package)
        .filter(model.Package.id == entity_id)
        .one_or_none()
    )
    if pkg:
        return pkg.name
    group = (
        model.Session.query(model.Group)
        .filter(model.Group.id == entity_id)
        .one_or_none()
    )
    if pkg:
        return group.name



import ckan.plugins as plugins
import ckan.plugins.toolkit as tk
from ckan.lib.search import rebuild
from ckan.logic import NotFound

import ckanext.scheming.helpers as sch

import ckanext.relationship.utils as utils


@tk.blanket.actions
@tk.blanket.auth_functions
@tk.blanket.validators
@tk.blanket.helpers
@tk.blanket.blueprints
class RelationshipPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IConfigurer)
    plugins.implements(plugins.IPackageController, inherit=True)

    # IConfigurer
    def update_config(self, config_):
        tk.add_template_directory(config_, "templates")
        tk.add_public_directory(config_, "public")
        tk.add_resource("assets", "relationship")

    # IPackageController
    def after_dataset_create(self, context, pkg_dict):
        context = context.copy()
        context.pop("__auth_audit", None)
        return _update_relations(context, pkg_dict)

    def after_dataset_update(self, context, pkg_dict):
        context = context.copy()
        context.pop("__auth_audit", None)
        return _update_relations(context, pkg_dict)

    def after_dataset_delete(self, context, pkg_dict):
        context = context.copy()
        context.pop("__auth_audit", None)

        subject_id = pkg_dict["id"]

        relations_ids_list = tk.get_action("relationship_relations_ids_list")(
            context, {"subject_id": subject_id}
        )

        for object_id in relations_ids_list:
            tk.get_action("relationship_relation_delete")(
                context, {"subject_id": subject_id, "object_id": object_id}
            )

            try:
                rebuild(object_id)
            except NotFound:
                pass
        rebuild(subject_id)

    def before_dataset_index(self, pkg_dict):
        pkg_id = pkg_dict["id"]
        pkg_type = pkg_dict["type"]
        schema = sch.scheming_get_schema("dataset", pkg_type)
        if not schema:
            return pkg_dict
        relations_info = utils.get_relations_info(pkg_type)
        for (
            related_entity,
            related_entity_type,
            relation_type,
        ) in relations_info:
            relations_ids = tk.get_action("relationship_relations_ids_list")(
                {},
                {
                    "subject_id": pkg_id,
                    "object_entity": related_entity,
                    "object_type": related_entity_type,
                    "relation_type": relation_type,
                },
            )

            if not relations_ids:
                continue
            field = utils.get_relation_field(
                pkg_type, related_entity, related_entity_type, relation_type
            )
            pkg_dict[f'vocab_{field["field_name"]}'] = relations_ids

            del pkg_dict[field["field_name"]]

        return pkg_dict

    def after_dataset_show(self, context, pkg_dict):
        pkg_id = pkg_dict["id"]
        pkg_type = pkg_dict["type"]
        relations_info = utils.get_relations_info(pkg_type)
        for (
            related_entity,
            related_entity_type,
            relation_type,
        ) in relations_info:
            field = utils.get_relation_field(
                pkg_type, related_entity, related_entity_type, relation_type
            )
            pkg_dict[field["field_name"]] = tk.get_action(
                "relationship_relations_ids_list"
            )(
                context,
                {
                    "subject_id": pkg_id,
                    "object_entity": related_entity,
                    "object_type": related_entity_type,
                    "relation_type": relation_type,
                },
            )


def _update_relations(context, pkg_dict):
    subject_id = pkg_dict["id"]
    add_relations = pkg_dict.get("add_relations", [])
    del_relations = pkg_dict.get("del_relations", [])
    if not add_relations and not del_relations:
        return pkg_dict
    for object_id, relation_type in del_relations + add_relations:
        if (object_id, relation_type) in add_relations:
            tk.get_action("relationship_relation_create")(
                context,
                {
                    "subject_id": subject_id,
                    "object_id": object_id,
                    "relation_type": relation_type,
                },
            )
        else:
            tk.get_action("relationship_relation_delete")(
                context,
                {
                    "subject_id": subject_id,
                    "object_id": object_id,
                    "relation_type": relation_type,
                },
            )

        try:
            rebuild(object_id)
        except NotFound:
            pass
    rebuild(subject_id)
    return pkg_dict



from flask import Blueprint, jsonify

import ckan.plugins.toolkit as tk


def get_blueprints():
    return [
        relationships,
    ]


relationships = Blueprint("relationships", __name__)


@relationships.route("/api/util/relationships/autocomplete")
def relationships_autocomplete():
    incomplete = tk.request.args.get("incomplete", "")
    entity_type = tk.request.args.get("entity_type", "dataset")
    updatable_only = tk.asbool(tk.request.args.get("updatable_only", "False"))
    packages = tk.get_action("package_search")(
        {},
        {
            "q": incomplete,
            "fq": f"type:{entity_type}",
            "fl": "id, title",
            "rows": 100,
            "include_private": True,
            "sort": "score desc",
        },
    )["results"]

    # packages = filter(
    #     lambda pkg: incomplete.lower() in pkg["title"].lower(), packages
    # )

    if updatable_only:
        packages = filter(
            lambda pkg: tk.h.check_access("package_update", {"id": pkg["id"]}), packages
        )

    result = {
        "ResultSet": {
            "Result": [
                {
                    "name": pkg["id"],
                    "title": pkg["title"],
                }
                for pkg in packages
            ]
        }
    }
    return jsonify(result)



from ckan.logic.schema import validator_args


@validator_args
def relation_create(not_empty, one_of, ignore_missing):
    return {
        "subject_id": [
            not_empty,
        ],
        "object_id": [
            not_empty,
        ],
        "relation_type": [
            one_of(["related_to", "child_of", "parent_of"]),
        ],
    }


@validator_args
def relation_delete(not_empty, one_of, ignore_missing):
    return {
        "subject_id": [
            not_empty,
        ],
        "object_id": [
            not_empty,
        ],
        "relation_type": [
            ignore_missing,
            one_of(["related_to", "child_of", "parent_of"]),
        ],
    }


@validator_args
def relations_list(not_empty, one_of, ignore_missing):
    return {
        "subject_id": [
            not_empty,
        ],
        "object_entity": [
            ignore_missing,
            one_of(["package", "organization", "group"]),
        ],
        "object_type": [
            ignore_missing,
        ],
        "relation_type": [
            ignore_missing,
            one_of(["related_to", "child_of", "parent_of"]),
        ],
    }


@validator_args
def relations_ids_list(not_empty, one_of, ignore_missing):
    return {
        "subject_id": [
            not_empty,
        ],
        "object_entity": [
            ignore_missing,
            one_of(["package", "organization", "group"]),
        ],
        "object_type": [
            ignore_missing,
        ],
        "relation_type": [
            ignore_missing,
            one_of(["related_to", "child_of", "parent_of"]),
        ],
    }


@validator_args
def get_entity_list(not_empty, one_of, ignore_missing):
    return {
        "entity": [
            not_empty,
            one_of(["package", "organization", "group"]),
        ],
        "entity_type": [
            not_empty,
        ],
    }



from __future__ import annotations

import json

from ckantoolkit import missing
from six import string_types

import ckan.plugins.toolkit as tk

from ckanext.scheming.validation import (
    scheming_multiple_choice_output,
    scheming_validator,
)


@scheming_validator
def relationship_related_entity(field, schema):
    related_entity = field.get("related_entity")
    related_entity_type = field.get("related_entity_type")
    relation_type = field.get("relation_type")

    def validator(key, data, errors, context):
        if field.get("required") and data[key] is missing:
            errors[key].append(tk._("Select at least one"))

        entity_id = data.get(("id",))

        current_relations = get_current_relations(
            entity_id, related_entity, related_entity_type, relation_type
        )

        selected_relations = get_selected_relations(data[key])

        data[key] = json.dumps([value for value in selected_relations])

        add_relations = selected_relations - current_relations
        del_relations = current_relations - selected_relations

        data[("add_relations",)] = data.get(("add_relations",), [])
        data[("del_relations",)] = data.get(("del_relations",), [])

        data[("add_relations",)].extend([(rel, relation_type) for rel in add_relations])
        data[("del_relations",)].extend([(rel, relation_type) for rel in del_relations])

    return validator


def get_current_relations(
    entity_id, related_entity, related_entity_type, relation_type
):
    if entity_id:
        current_relations = tk.get_action("relationship_relations_list")(
            {},
            {
                "subject_id": entity_id,
                "object_entity": related_entity,
                "object_type": related_entity_type,
                "relation_type": relation_type,
            },
        )
        current_relations = [rel["object_id"] for rel in current_relations]
    else:
        current_relations = []
    return set(current_relations)


def get_selected_relations(selected_relations):
    if isinstance(selected_relations, string_types) and "," in selected_relations:
        selected_relations = selected_relations.split(",")

    if (
        len(selected_relations) == 1
        and isinstance(selected_relations[0], string_types)
        and "," in selected_relations[0]
    ):
        selected_relations = selected_relations[0].split(",")

    if selected_relations is not missing:
        selected_relations = scheming_multiple_choice_output(selected_relations)
        selected_relations = [] if selected_relations == [""] else selected_relations
    else:
        selected_relations = []
    return set(selected_relations)



from __future__ import annotations

from ckan.plugins import toolkit as tk


def relationship_relation_create(context, data_dict):
    return {"success": True}


def relationship_relation_delete(context, data_dict):
    return {"success": True}


@tk.auth_allow_anonymous_access
def relationship_relations_list(context, data_dict):
    return {"success": True}


@tk.auth_allow_anonymous_access
def relationship_relations_ids_list(context, data_dict):
    return {"success": True}


@tk.auth_allow_anonymous_access
def relationship_get_entity_list(context, data_dict):
    return {"success": True}



# -*- coding: utf-8 -*-
# Always prefer setuptools over distutils
from codecs import open  # To use a consistent encoding
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="""ckanext-relationship""",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # http://packaging.python.org/en/latest/tutorial.html#version
    version="0.2.2",
    description="""""",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # The project's main homepage.
    url="https://github.com/DataShades/ckanext-relationship",
    # Author details
    author="""Oleksandr Ivaniuk""",
    author_email="""aleks.ivaniuk@gmail.com""",
    # Choose your license
    license="AGPL",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        # 3 - Alpha
        # 4 - Beta
        # 5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Pick your license as you wish (should match "license" above)
        (
            "License :: OSI Approved :: GNU Affero General Public License v3"
            " or later (AGPLv3+)"
        ),
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 2.7",
    ],
    # What does your project relate to?
    keywords="""CKAN""",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=["contrib", "docs", "tests*"]),
    namespace_packages=["ckanext"],
    install_requires=[
        # CKAN extensions should not list dependencies here, but in a separate
        # ``requirements.txt`` file.
        #
        # http://docs.ckan.org/en/latest/extensions/best-practices.html
        # add-third-party-libraries-to-requirements-txt
    ],
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    include_package_data=True,
    package_data={},
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages.
    # see http://docs.python.org/3.4/distutils/setupscript.html
    # installing-additional-files
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[],
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points="""
        [ckan.plugins]
        relationship=ckanext.relationship.plugin:RelationshipPlugin

        [babel.extractors]
        ckan = ckan.lib.extract:extract_ckan
    """,
    # If you are changing from the default layout of your extension, you may
    # have to change the message extractors, you can read more about babel
    # message extraction at
    # http://babel.pocoo.org/docs/messages/#extraction-method-mapping-and-configuration
    message_extractors={
        "ckanext": [
            ("**.py", "python", None),
            ("**.js", "javascript", None),
            ("**/templates/**.html", "ckan", None),
        ],
    },
)
